@@ -1,23 +1,8 @@
 /*
+ * Embedded Linux library
+ * Copyright (C) 2015-2019  Intel Corporation
  *
- *  Embedded Linux library
- *
- *  Copyright (C) 2015-2019  Intel Corporation. All rights reserved.
- *
- *  This library is free software; you can redistribute it and/or
- *  modify it under the terms of the GNU Lesser General Public
- *  License as published by the Free Software Foundation; either
- *  version 2.1 of the License, or (at your option) any later version.
- *
- *  This library is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- *  Lesser General Public License for more details.
- *
- *  You should have received a copy of the GNU Lesser General Public
- *  License along with this library; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
- *
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 #ifdef HAVE_CONFIG_H
@@ -363,6 +348,84 @@ static void test_uintset_isempty(const void *user_data)
 	l_uintset_free(b);
 }
 
+static void test_uintset_size(const void *user_data)
+{
+	uint32_t i;
+	struct l_uintset *set = l_uintset_new(32);
+
+	assert(l_uintset_size(set) == 0);
+
+	for (i = 1; i < 32; i++) {
+		assert(l_uintset_put(set, i));
+		assert(l_uintset_size(set) == i);
+	}
+
+	assert(l_uintset_size(set) == i - 1);
+
+	assert(l_uintset_take(set, 10));
+	assert(l_uintset_size(set) == i - 2);
+
+	for (i = 1; i < 32; i++)
+		assert(l_uintset_take(set, i));
+
+	assert(l_uintset_size(set) == 0);
+
+	l_uintset_free(set);
+}
+
+static void test_uintset_subtract(const void *data)
+{
+	struct l_uintset *set_a = l_uintset_new_from_range(0, 10);
+	struct l_uintset *set_b = l_uintset_new_from_range(1, 11);
+	struct l_uintset *sub;
+
+	/* Some sanity checks */
+	assert(l_uintset_subtract(NULL, NULL) == NULL);
+	assert(l_uintset_subtract(set_a, set_b) == NULL);
+
+	l_uintset_free(set_a);
+	l_uintset_free(set_b);
+
+	set_a = l_uintset_new_from_range(0, 128);
+	set_b = l_uintset_new_from_range(0, 128);
+
+	/* (<empty>) - (<empty>) = (<empty>) */
+	sub = l_uintset_subtract(set_a, set_b);
+	assert(l_uintset_isempty(sub));
+	l_uintset_free(sub);
+
+	/* Sanity check this works across word boundaries */
+	l_uintset_put(set_b, 65);
+	l_uintset_put(set_b, 64);
+	l_uintset_put(set_b, 63);
+
+	/* (<empty>) - (65, 64, 63) = (<empty>) */
+	sub = l_uintset_subtract(set_a, set_b);
+	assert(!l_uintset_contains(sub, 64));
+	l_uintset_free(sub);
+
+	l_uintset_put(set_a, 64);
+
+	/* (64) - (65, 64, 63) = (<empty>) */
+	sub = l_uintset_subtract(set_a, set_b);
+	assert(!l_uintset_contains(sub, 64));
+	l_uintset_free(sub);
+
+	l_uintset_take(set_b, 64);
+
+	/* (64) - (65, 63) = (64) */
+	sub = l_uintset_subtract(set_a, set_b);
+	assert(l_uintset_contains(sub, 64));
+	l_uintset_free(sub);
+
+	/* (65, 63) - (64) = (65, 63) */
+	sub = l_uintset_subtract(set_b, set_a);
+	assert(l_uintset_contains(sub, 65));
+	assert(!l_uintset_contains(sub, 64));
+	assert(l_uintset_contains(sub, 63));
+	l_uintset_free(sub);
+}
+
 int main(int argc, char *argv[])
 {
 	l_test_init(&argc, &argv);
@@ -381,6 +444,8 @@ int main(int argc, char *argv[])
 	l_test_add("l_uintset intersect test 2", test_uintset_intersect_test,
 							&intersect_data_2);
 	l_test_add("l_uintset isempty", test_uintset_isempty, NULL);
+	l_test_add("l_uintset size", test_uintset_size, NULL);
+	l_test_add("l_uintset_subtract", test_uintset_subtract, NULL);
 
 	return l_test_run();
 }
